@@ -16,7 +16,6 @@ import logging
 import traceback
 import asyncio
 import logging
-import nest_asyncio
 
 from typing import Any, Dict, List
 from acryl.executor.common.config import ConfigModel
@@ -49,17 +48,13 @@ Executes a Task Execution Request
 class DefaultExecutor(Executor):
     id: str 
 
-    # Execution Event Loop
-    event_loop = asyncio.new_event_loop()
-
-    # Patch the primary event loop to permit multiple runs 
-    nest_asyncio.apply(event_loop)
-
-
     # Tasks
     task_registry: TaskRegistry = TaskRegistry()
     task_instances: Dict[str, Task] = {}
-    task_futures: Dict[str, List[Any]] = {}
+
+    # Execution event loops + futures.
+    task_event_loops: Dict[str, asyncio.AbstractEventLoop] = {}
+    task_futures: Dict[str, List[asyncio.Task]] = {}
 
     # Secret Stores 
     secret_stores: List[SecretStore] = []
@@ -71,9 +66,6 @@ class DefaultExecutor(Executor):
         for task_config in config.task_configs:
             self.register_task(task_config)
 
-        # Set event loop context 
-        asyncio.set_event_loop(self.event_loop)
-
     def register_task(self, task: TaskConfig) -> None:
         # Store a registry of command -> task. 
         self.task_registry.register_lazy(task.name, task.type)
@@ -97,6 +89,7 @@ class DefaultExecutor(Executor):
         return self.execute_task(request, executor_context)
 
     def execute_task(self, request: ExecutionRequest, executor_context: ExecutorContext) -> ExecutionResult:
+        # ASSUMPTION: Each time execute_task is called, it is called from a different thread.
 
         # 1. Build task execution context
         execution_context = ExecutionContext(request)
@@ -104,6 +97,8 @@ class DefaultExecutor(Executor):
         execution_report = execution_context.get_report()
         execution_report.report_info(f"Starting execution for task with name={request.name}")
 
+        assert execution_id not in self.task_futures, "Already running task with same execution ID"
+
         # 2. Retrieve an instance of the task  
         task_instance = self.task_instances.get(request.name)
         execution_result = ExecutionResult(execution_context)
@@ -114,12 +109,18 @@ class DefaultExecutor(Executor):
 
         # 3. Execute task - task is expected to throw TaskError on failure. 
         try:
-            # Execute task via an event loop. Store the executing task such that it can be cancelled. 
+            # 3.1. Setup event loop for task execution.
+            task_event_loop = asyncio.new_event_loop()
+            self.task_event_loops[execution_id] = task_event_loop
+            asyncio.set_event_loop(task_event_loop)
+
+            # 3.2. Execute task via the thread's event loop. Store the executing task such that it can be cancelled. 
             # TODO: Have an async version of this method. 
-            task_future = self.event_loop.create_task(task_instance.execute(request.args, execution_context))
+            task_future = task_event_loop.create_task(task_instance.execute(request.args, execution_context))
+            task_future.set_name(f"dispatch-{execution_id}")
             logger.debug(f"Task for {execution_id} created")
             self.task_futures[execution_id] = [task_future]
-            self.event_loop.run_until_complete(task_future)
+            task_event_loop.run_until_complete(task_future)
             logger.debug(f"Task for {execution_id} completed")
         except Exception:
             execution_report.report_info(f"Caught exception EXECUTING task_id={execution_context.get_execution_id()}, name={request.name}, stacktrace={traceback.format_exc(limit=20)}")
@@ -128,6 +129,12 @@ class DefaultExecutor(Executor):
             execution_report.report_info(f"Execution cancelled while EXECUTING task_id={execution_context.get_execution_id()}, name={request.name}, stacktrace={traceback.format_exc(limit=3)}")
             execution_result.set_result_type(Type.CANCELLED)
             return execution_result 
+        finally:
+            # task_event_loop.close()
+            # del self.task_futures[execution_id]
+            # del self.task_event_loops[execution_id]
+
+            logger.debug(f"Cleaned up task for {execution_id}")
 
         execution_result.set_result_type(Type.SUCCESS)
         return execution_result
@@ -138,12 +145,12 @@ class DefaultExecutor(Executor):
             if task_futures is None:
                 logger.error(f"Received KILL for missing task ID {request.exec_id}")
             else:
-                for task_future in task_futures: 
+                event_loop = self.task_event_loops[request.exec_id]
+                for task_future in task_futures:
                     # Cancel the task if not complete. 
                     if not task_future.done():
                         logger.debug(f"Trying to cancel {task_future}")
-                        task_future.cancel()
-                
+                        event_loop.call_soon_threadsafe(task_future.cancel)
 
     def shutdown(self) -> None:
         for task_name, task_instance in self.task_instances.items():
@@ -178,4 +185,4 @@ class DefaultExecutor(Executor):
         secret_store_class = secret_store_registry.get(secret_store_type)
 
         # Create & return new instance of secret store 
-        return secret_store_class.create(config.config)
\ No newline at end of file
+        return secret_store_class.create(config.config)
